from .cli import main

__name__ = "curlylint"
__version__ = "0.5.0"
__description__ = "A linter for Jinja-like templates. Forked from jinjalint"
__author__ = "Thibaud Colas"
__author_email__ = "thibaudcolas@gmail.com"
__url__ = "https://github.com/thibaudcolas/curlylint"
__license__ = "MIT"
__copyright__ = "Copyright 2020-present Thibaud Colas"

if __name__ == '__main__':
    main()
